#if !defined (UNITTEST)
#include "globals.h"
#else
#include "StdAfx.h"
#endif
#include "RuleTreeNode.h"
#include "keymap.h"
#include "jstring.h"

CSkkRuleTreeNodeOutput*
CSkkRuleTreeNodeOutputStringPair::pCreateInstance (
	LPCDSTR							dstrLeft,
	int								nLeftLen,
	LPCDSTR							dstrRight,
	int								nRightLen)
{
	CSkkRuleTreeNodeOutputStringPair*	pNode ;
	DCHAR*	ptr ;

	pNode	= new CSkkRuleTreeNodeOutputStringPair () ;
	if (pNode == NULL)
		return	NULL ;
	if (nLeftLen > 0 && dstrLeft != NULL) {
		ptr	= new DCHAR [nLeftLen] ;
		if (ptr == NULL) {
			delete	pNode ;
			return	NULL ;
		}
		memcpy (ptr, dstrLeft, nLeftLen * sizeof (DCHAR)) ;
		pNode->m_strHira	= ptr ;
		pNode->m_nHira		= nLeftLen ;
	}
	if (nRightLen > 0 && dstrRight != NULL) {
		ptr	= new DCHAR [nRightLen] ;
		if (ptr == NULL) {
			delete	pNode ;
			return	NULL ;
		}
		memcpy (ptr, dstrRight, nRightLen * sizeof (DCHAR)) ;
		pNode->m_strKata	= ptr ;
		pNode->m_nKata		= nRightLen ;
	}
	return	pNode ;
}

CSkkRuleTreeNodeOutput*
CSkkRuleTreeNodeOutputString::pCreateInstance (
	LPCDSTR							dstrString,
	int								nStringLen)
{
	CSkkRuleTreeNodeOutputString*	pNode ;
	DCHAR*	ptr ;

	pNode	= new CSkkRuleTreeNodeOutputString () ;
	if (pNode == NULL)
		return	NULL ;
	if (dstrString != NULL && nStringLen > 0) {
		ptr	= new DCHAR [nStringLen] ;
		if (ptr == NULL) {
			delete	pNode ;
			return	NULL ;
		}
		memcpy (ptr, dstrString, nStringLen * sizeof (DCHAR)) ;
		pNode->m_strValue	= ptr ;
		pNode->m_nValue		= nStringLen ;
	} 
	return	pNode ;
}

CSkkRuleTreeNodeOutput*
CSkkRuleTreeNodeOutputMacro::pCreateInstance (int nFuncNo) 
{
	CSkkRuleTreeNodeOutputMacro*	pNode ;

	pNode	= new CSkkRuleTreeNodeOutputMacro () ;
	if (pNode == NULL)
		return	NULL ;
	pNode->m_nMacro	= nFuncNo ;
	return	pNode ;
}

/*===============================================================
 *	static public methods
 */
LPCDSTR
CSkkRuleTreeNode::pGetPrefix (
	const CSkkRuleTreeNode*			pNode,
	int*							pnLength)
{
	if (pNode == NULL)
		return	NULL ;
	if (pnLength != NULL)
		*pnLength	= pNode->m_nPrefix ;
	return	pNode->m_pPrefix ;
}

LPCDSTR
CSkkRuleTreeNode::pGetPrefix (
	CSkkRuleTreeNode*				pNode,
	int*							pnLength)
{
	if (pNode == NULL)
		return	NULL ;
	if (pnLength != NULL)
		*pnLength	= pNode->m_nPrefix ;
	return	pNode->m_pPrefix ;
}

CSkkRuleTreeNode*
CSkkRuleTreeNode::pCreateInstance (
	int								iRule,
	int								nChar, 
	LPCDSTR							dstrPrefix,
	int								nPrefixLen,
	LPCDSTR							dstrNext,
	int								nNextLen,
	int								nNextRule,
	CSkkRuleTreeNodeOutput*			pOutput)
{
	CSkkRuleTreeNode*	pNode ;
	DCHAR*				pwDest ;

	pNode	= new CSkkRuleTreeNode () ;
	if (pNode == NULL)
		return	NULL ;

	pNode->m_iRule	= iRule ;
	pNode->m_nCH	= nChar ;
	if (nPrefixLen > 0 && dstrPrefix != NULL) {
		pwDest		= new DCHAR [nPrefixLen] ;
		if (pwDest == NULL) {
			delete	pNode ;
			return	NULL ;
		}
		pNode->m_pPrefix	= pwDest ;
		memcpy (pNode->m_pPrefix, dstrPrefix, sizeof (DCHAR) * nPrefixLen) ;
	} else {
		nPrefixLen		= 0 ;
	}
	pNode->m_nPrefix	= nPrefixLen ;

	if (nNextLen > 0 && dstrNext != NULL) {
		pwDest		= new DCHAR [nNextLen] ;
		pNode->m_pNextState	= pwDest ;
		memcpy (pNode->m_pNextState, dstrNext, sizeof (DCHAR) * nNextLen) ;
	} else {
		nNextLen	= 0 ;
	}
	pNode->m_nNextState	= nNextLen ;
	pNode->m_nNextRule	= nNextRule ;
	pNode->m_pOutput	= pOutput ;
	return	pNode ;
}

CSkkRuleTreeNode*
CSkkRuleTreeNode::pGetBranchList (
	CSkkRuleTreeNode*		pRoot)
{
	return	pRoot != NULL? pRoot->pGetChild () : NULL ;
}

const CSkkRuleTreeNode*
CSkkRuleTreeNode::pGetBranchList (
	const CSkkRuleTreeNode*		pRoot)
{
	return	pRoot != NULL? pRoot->pGetChild () : NULL ;
}

BOOL
CSkkRuleTreeNode::bMatchChar (int nSystemChar, int nInputChar, int* pnScore)
{
	switch (nSystemChar) {
	case	CSkkRuleTreeNode::SYSCHAR_DEFAULT:
		if (pnScore != NULL)
			*pnScore	= 0 ;	/* Œ_B*/
		return	TRUE ;
	case	CSkkRuleTreeNode::SYSCHAR_VOWEL:
		if	(nInputChar == 'a' || nInputChar == 'i' || nInputChar == 'u' || nInputChar == 'e' || nInputChar == 'o') {
			if (pnScore != NULL)
				*pnScore	= 50 ;
			return	TRUE ;
		}
		break ;
	case	CSkkRuleTreeNode::SYSCHAR_CONSONANT:
		if	(('a' <= nInputChar && nInputChar <= 'z') && ! (nInputChar == 'a' || nInputChar == 'i' || nInputChar == 'u' || nInputChar == 'e' || nInputChar == 'o')) {
			if (pnScore != NULL)
				*pnScore	= 50 ;
			return	TRUE ;
		}
		break ;
	default:
		if (nSystemChar == nInputChar) {
			if (pnScore != NULL)
				*pnScore	= 100 ;
			return	TRUE ;
		}
		break ;
	}
	return	FALSE ;
}

CSkkRuleTreeNode*
CSkkRuleTreeNode::pSelectBranch (
	CSkkRuleTreeNode*		pRoot,
	int						nChar)
{
	CSkkRuleTreeNode*	pNode ;
	CSkkRuleTreeNode*	pHitNode ;
	int					nCurScore ;

	if (pRoot == NULL)
		return	NULL ;

	pNode		= pRoot->pGetChild () ;
	pHitNode	= NULL ;
	nCurScore	= -1 ;
	while (pNode != NULL) {
		int	nScore ;
		if (CSkkRuleTreeNode::bMatchChar (pNode->iGetChar (), nChar, &nScore)) {
			if (nCurScore < nScore) {
				pHitNode	= pNode ;
				nCurScore	= nScore ;
			}
			if (nScore == 100)
				break ;
		}
		pNode	= pNode->pGetBrother () ;
	}
	return	pHitNode ;
}

const CSkkRuleTreeNode*
CSkkRuleTreeNode::pSelectBranch (
	const CSkkRuleTreeNode*		pRoot,
	int							nChar)
{
	const CSkkRuleTreeNode*	pNode ;
	const CSkkRuleTreeNode*	pHitNode ;
	int	nCurScore ;

	if (pRoot == NULL)
		return	NULL ;

	pNode		= pRoot->pGetChild () ;
	pHitNode	= NULL ;
	nCurScore	= -1 ;
	while (pNode != NULL) {
		int	nScore ;
		if (CSkkRuleTreeNode::bMatchChar (pNode->iGetChar (), nChar, &nScore)) {
			if (nCurScore < nScore) {
				pHitNode	= pNode ;
				nCurScore	= nScore ;
			}
			if (nScore == 100)
				break ;
		}
		pNode	= pNode->pGetBrother () ;
	}
	return	pHitNode ;
}

/*	ԂĺATɐ(ǂȂȂ炻܂)
 *	̎_ł̃m[h ppResult ɓB
 */
int
CSkkRuleTreeNode::iSearchTree (
	CSkkRuleTreeNode*			pTree,
	LPCDSTR						pdString,
	int							nString,
	CSkkRuleTreeNode**			ppResult)
{
	CSkkRuleTreeNode*	pParent ;
	CSkkRuleTreeNode*	pNode ;
	LPCDSTR		wptr, wptrEnd ;
	int			nCH ;

	pNode		= pTree != NULL? pTree->pGetChild () : NULL ;
	pParent		= pTree ;
	wptr		= pdString ;
	wptrEnd		= pdString + nString ;
	while (wptr < wptrEnd) {
		LPCDSTR	wptrBack	= wptr ;
		nCH		= CSkkRuleTreeNode::iGetRuleCharacter (&wptr, wptrEnd) ;
		/* Z푤̒TB*/
		while (pNode != NULL) {
			/*	[̈vB
			 */
			if (pNode->iGetChar () == nCH)
				break ;
			pNode	= pNode->pGetBrother () ;
		}
		if (pNode == NULL) {
			/* łBvĂƂ܂ŕԂB*/
			if (ppResult != NULL)
				*ppResult	= pParent ;
#if 0 //defined (DEBUG)
			DEBUGPRINTF ((TEXT ("iSeachTree: not found\n"))) ;
#endif
			return	wptrBack - pdString ;
		}
		if (wptr >= wptrEnd) {
			/* B*/
			if (ppResult != NULL)
				*ppResult	= pNode ;
#if 0 //defined (DEBUG)
			DEBUGPRINTF ((TEXT ("iSeachTree: found\n"))) ;
			pNode->vDebugOut () ;
			pParent->vDebugOut () ;
#endif
			return	wptr - pdString ;
		}
#if 0 //defined (DEBUG)
		DEBUGPRINTF ((TEXT ("iSearchTree: -> goto child\n"))) ;
		pNode->vDebugOut () ;
#endif
		pParent	= pNode ;
		pNode	= pNode->pGetChild () ;
	}
	if (ppResult != NULL)
		*ppResult	= pTree ;
	return	0 ;
}

BOOL
CSkkRuleTreeNode::bAddRule (
	CSkkRuleTreeNode**	ppTree,
	int					iTree,
	LPCDSTR				dstrPrefix,
	int					nPrefixLen,
	LPCWSTR				wstrNext,
	int					nNextLen,
	int					nNextTree,
	LPCWSTR				wstrHira,
	int					nHiraLen,
	LPCWSTR				wstrKata,
	int					nKataLen)
{
	CSkkRuleTreeNode*		pRoot ;
	CSkkRuleTreeNode*		pAddPoint		= NULL ;
	CSkkRuleTreeNode*		pAddTree ;
	CSkkRuleTreeNode*		pAddTreeRoot	= NULL ;
	CSkkRuleTreeNodeOutput*	pOutput			= NULL ;
	LPCDSTR					pwPtr ;
	LPCDSTR					pwPtrEnd ;
	int						nCH, nOffset ;
	DCHAR					bufHira [32], bufKata [32], bufNext [32] ;
	DCHAR*					pdHira = NULL ;
	DCHAR*					pdKata = NULL ;
	DCHAR*					pdNext	= NULL ;
	int						nKataDLen, nHiraDLen, nNextDLen ;
	BOOL					bRetval	= TRUE ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || dstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= pCreateInstance (iTree, 0, NULL, 0, NULL, 0, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}

	nOffset		= CSkkRuleTreeNode::iSearchTree (pRoot, dstrPrefix, nPrefixLen, &pAddPoint) ;

	pwPtr		= dstrPrefix + nOffset ;
	pwPtrEnd	= dstrPrefix + nPrefixLen ;
	if (pwPtr < pwPtrEnd) {
		CSkkRuleTreeNode*		pCurAddPoint		= NULL ;

		/*	S̐VKo^łꍇB*/
		/*
		 *	shu ƂāAs ݂ꍇɂ nOffset = 1B
		 *
		 *	[g -> ``s'' -> ...
		 *
		 *	 ``s''  pAddPoint ƂČtĂBA
		 *
		 *	(1) ``h'' -> ``u'' ̃N쐬B
		 *	(2) ``s'' ̎qƂĒǉB
		 *		ɂ ``s'' -> ``h'' -> ``u''  rule-tree 쐬B
		 *
		 *	 ``s'' ̎q݂ꍇɂ́A
		 *
		 *	(2)' ``h''->``u''  ``h'' ̌ZɂƂƂ ``s'' ̎qǉB
		 *	(3)  ``s'' ̎qƂ ``h'' ȂB
		 *
		 *	̃[ɈᔽȂ悤Ƀ[؂̍\gȂ΂ȂȂB
		 *	nCH  INT Ȃ̂ŁA̕ԈႦȂΊgɎs邱Ƃ͂Ȃ
		 *	̂ȂB
		 *	nCH ɓꕶꂽꍇ́uݏԂ̊ǗvB͂ǂɂ
		 *	Ȃ΂ȂȂB̓[؂̒ɓ̂͂܂ȂB
		 *	nCH ̓ꕶg
		 *		SelectBranch
		 *	͕ωB
		 *	Ԃ m_pPrefix, m_nPrefix A̎ۂۑꏊKvB
		 *	ۂ̏Ԃ́AiCH, m_pPrefix, m_nPrefix 쐬锤B
		 */
		pAddTreeRoot	= NULL ;
		while (pwPtr < pwPtrEnd) {
			LPCDSTR	pwPtrBack	= pwPtr ;

			nCH			= CSkkRuleTreeNode::iGetRuleCharacter (&pwPtr, pwPtrEnd) ;
			if (nCH == CSkkRuleTreeNode::SYSCHAR_INVALID) {
				return	FALSE ;
			}
			if (pwPtr >= pwPtrEnd) {
				pwPtr	= pwPtrBack ;
				break ;
			}
			pAddTree	= CSkkRuleTreeNode::pCreateInstance (iTree, nCH, dstrPrefix, pwPtr - dstrPrefix, NULL, 0, iTree, NULL) ;
			if (pAddTree == NULL) {
				vDestroyTree (pAddTreeRoot) ;
				return	FALSE ;
			}

			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			/*	addpoint  addtree ǉB\[gȂȂycB*/
			if (pAddTreeRoot == NULL) {
				pAddTreeRoot	= pAddTree ;
				pCurAddPoint	= pAddTree ;
			} else {
				pAddTree->vSetBrother (pCurAddPoint->pGetChild ()) ;
				pCurAddPoint->vSetChild (pAddTree) ;
				pCurAddPoint	= pAddTree ;
			}
		}
		pdHira	= pGetDString (bufHira, MYARRAYSIZE (bufHira), wstrHira, nHiraLen, &nHiraDLen) ;
		pdKata	= pGetDString (bufKata, MYARRAYSIZE (bufKata), wstrKata, nKataLen, &nKataDLen) ;

		/*	Output 쐬B
		 */
		pOutput		= CSkkRuleTreeNodeOutputStringPair::pCreateInstance (pdHira, nHiraDLen, pdKata, nKataDLen) ;
		if (pOutput == NULL) {
			vDestroyTree (pAddTreeRoot) ;
			bRetval	= FALSE ;
			goto	Exit ;
		}

		/*	Output ̂Ȃ RuleTree ̃m[h쐬B
		 */
		nCH	= CSkkRuleTreeNode::iGetRuleCharacter (&pwPtr, pwPtrEnd) ;
		if (nCH == CSkkRuleTreeNode::SYSCHAR_INVALID) {
			vDestroyTree (pAddTreeRoot) ;
			delete	pOutput ;
			bRetval	= FALSE ;
			goto	Exit ;
		}
		pdNext	= pGetDString (bufNext, MYARRAYSIZE (bufNext), wstrNext, nNextLen, &nNextDLen) ;

		pAddTree	= CSkkRuleTreeNode::pCreateInstance (iTree, nCH, dstrPrefix, nPrefixLen, pdNext, nNextDLen, nNextTree, pOutput) ;
		if (pAddTree == NULL) {
			vDestroyTree (pAddTreeRoot) ;
			delete	pOutput ;
			bRetval	= FALSE ;
			goto	Exit ;
		}
		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		if (pAddTreeRoot != NULL) {
			ASSERT (pCurAddPoint != NULL && pCurAddPoint != pAddPoint) ;
			pAddTree->vSetBrother (pCurAddPoint->pGetChild ()) ;
			pCurAddPoint->vSetChild (pAddTree) ;
		} else {
			pAddTreeRoot	= pAddTree ;
		}

		/*	̒iKłƂ̖؂Ƀ}[WB
		 */
		pAddTreeRoot->vSetBrother (pAddPoint->pGetChild ()) ;
		pAddPoint->vSetChild (pAddTreeRoot) ;
	} else {
		pdHira	= pGetDString (bufHira, MYARRAYSIZE (bufHira), wstrHira, nHiraLen, &nHiraDLen) ;
		pdKata	= pGetDString (bufKata, MYARRAYSIZE (bufKata), wstrKata, nKataLen, &nKataDLen) ;

		/*	NextState ̍XVȂƂȂ̂ł́B
		 */
		pOutput		= CSkkRuleTreeNodeOutputStringPair::pCreateInstance (pdHira, nHiraDLen, pdKata, nKataDLen) ;
		if (pOutput == NULL) {
			bRetval	= FALSE ;
			goto	Exit ;
		}
		pAddPoint->vSetOutput (pOutput) ;
	}
Exit:
	if (pdHira != bufHira)
		delete[]	pdHira ;
	if (pdKata != bufKata)
		delete[]	pdHira ;
	if (pdNext != bufNext)
		delete[]	pdNext ;
	return	bRetval ;
}

BOOL
CSkkRuleTreeNode::bAddRule (
	CSkkRuleTreeNode**	ppTree,
	int					iTree,
	LPCDSTR				dstrPrefix,
	int					nPrefixLen,
	LPCWSTR				wstrNext,
	int					nNextLen,
	int					nNextTree,
	LPCWSTR				wstrString,
	int					nString)
{
	CSkkRuleTreeNode*		pRoot ;
	CSkkRuleTreeNode*		pAddPoint		= NULL ;
	CSkkRuleTreeNode*		pAddTree ;
	CSkkRuleTreeNode*		pAddTreeRoot	= NULL ;
	CSkkRuleTreeNodeOutput*	pOutput			= NULL ;
	LPCDSTR					pwPtr ;
	LPCDSTR					pwPtrEnd ;
	int						nCH, nOffset ;
	DCHAR					bufString [32], bufNext [32] ;
	DCHAR*					pdString = NULL ;
	DCHAR*					pdNext	= NULL ;
	int						nStringDLen, nNextDLen ;
	BOOL					bRetval	= TRUE ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || dstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= pCreateInstance (iTree, 0, NULL, 0, NULL, 0, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}

	nOffset		= CSkkRuleTreeNode::iSearchTree (pRoot, dstrPrefix, nPrefixLen, &pAddPoint) ;

	pwPtr		= dstrPrefix + nOffset ;
	pwPtrEnd	= dstrPrefix + nPrefixLen ;
	if (pwPtr < pwPtrEnd) {
		CSkkRuleTreeNode*		pCurAddPoint		= NULL ;

		/*	S̐VKo^łꍇB*/
		/*
		 *	shu ƂāAs ݂ꍇɂ nOffset = 1B
		 *
		 *	[g -> ``s'' -> ...
		 *
		 *	 ``s''  pAddPoint ƂČtĂBA
		 *
		 *	(1) ``h'' -> ``u'' ̃N쐬B
		 *	(2) ``s'' ̎qƂĒǉB
		 *		ɂ ``s'' -> ``h'' -> ``u''  rule-tree 쐬B
		 *
		 *	 ``s'' ̎q݂ꍇɂ́A
		 *
		 *	(2)' ``h''->``u''  ``h'' ̌ZɂƂƂ ``s'' ̎qǉB
		 *	(3)  ``s'' ̎qƂ ``h'' ȂB
		 *
		 *	̃[ɈᔽȂ悤Ƀ[؂̍\gȂ΂ȂȂB
		 *	nCH  INT Ȃ̂ŁA̕ԈႦȂΊgɎs邱Ƃ͂Ȃ
		 *	̂ȂB
		 *	nCH ɓꕶꂽꍇ́uݏԂ̊ǗvB͂ǂɂ
		 *	Ȃ΂ȂȂB̓[؂̒ɓ̂͂܂ȂB
		 *	nCH ̓ꕶg
		 *		SelectBranch
		 *	͕ωB
		 *	Ԃ m_pPrefix, m_nPrefix A̎ۂۑꏊKvB
		 *	ۂ̏Ԃ́AiCH, m_pPrefix, m_nPrefix 쐬锤B
		 */
		pAddTreeRoot	= NULL ;
		while (pwPtr < pwPtrEnd) {
			LPCDSTR	pwPtrBack	= pwPtr ;

			nCH			= CSkkRuleTreeNode::iGetRuleCharacter (&pwPtr, pwPtrEnd) ;
			if (nCH == CSkkRuleTreeNode::SYSCHAR_INVALID) {
				return	FALSE ;
			}
			if (pwPtr >= pwPtrEnd) {
				pwPtr	= pwPtrBack ;
				break ;
			}

			pAddTree	= CSkkRuleTreeNode::pCreateInstance (iTree, nCH, dstrPrefix, pwPtr - dstrPrefix, NULL, 0, iTree, NULL) ;
			if (pAddTree == NULL) {
				vDestroyTree (pAddTreeRoot) ;
				bRetval	= FALSE ;
				goto	Exit ;
			}

			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			/*	addpoint  addtree ǉB\[gȂȂycB*/
			if (pAddTreeRoot == NULL) {
				pAddTreeRoot	= pAddTree ;
				pCurAddPoint	= pAddTree ;
			} else {
				pAddTree->vSetBrother (pCurAddPoint->pGetChild ()) ;
				pCurAddPoint->vSetChild (pAddTree) ;
				pCurAddPoint	= pAddTree ;
			}
		}

		/*	Output 쐬B
		 */
		pdString	= pGetDString (bufString, MYARRAYSIZE (bufString), wstrString, nString, &nStringDLen) ;
		pOutput		= CSkkRuleTreeNodeOutputString::pCreateInstance (pdString, nStringDLen) ;
		if (pOutput == NULL) {
			vDestroyTree (pAddTreeRoot) ;
			bRetval	= FALSE ;
			goto	Exit ;
		}

		/*	Output ̂Ȃ RuleTree ̃m[h쐬B
		 */
		nCH	= CSkkRuleTreeNode::iGetRuleCharacter (&pwPtr, pwPtrEnd) ;
		if (nCH == CSkkRuleTreeNode::SYSCHAR_INVALID) {
			vDestroyTree (pAddTreeRoot) ;
			delete	pOutput ;
			return	FALSE ;
		}
		pdNext		= pGetDString (bufNext, MYARRAYSIZE (bufNext), wstrNext, nNextLen, &nNextDLen) ;
		pAddTree	= CSkkRuleTreeNode::pCreateInstance (iTree, nCH, dstrPrefix, nPrefixLen, pdNext, nNextDLen, nNextTree, pOutput) ;
		if (pAddTree == NULL) {
			vDestroyTree (pAddTreeRoot) ;
			delete	pOutput ;
			bRetval	= FALSE ;
			goto	Exit ;
		}
		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		if (pAddTreeRoot != NULL) {
			ASSERT (pCurAddPoint != NULL && pCurAddPoint != pAddPoint) ;
			pAddTree->vSetBrother (pCurAddPoint->pGetChild ()) ;
			pCurAddPoint->vSetChild (pAddTree) ;
		} else {
			pAddTreeRoot	= pAddTree ;
		}

		/*	̒iKłƂ̖؂Ƀ}[WB
		 */
		pAddTreeRoot->vSetBrother (pAddPoint->pGetChild ()) ;
		pAddPoint->vSetChild (pAddTreeRoot) ;
	} else {
		pdString	= pGetDString (bufString, MYARRAYSIZE (bufString), wstrString, nString, &nStringDLen) ;
		pOutput		= CSkkRuleTreeNodeOutputString::pCreateInstance (pdString, nStringDLen) ;
		if (pOutput == NULL) {
			bRetval	= FALSE ;
			goto	Exit ;
		}
		pAddPoint->vSetOutput (pOutput) ;
	}
Exit:
	if (pdNext != bufNext)
		delete[]	pdNext ;
	if (pdString != bufString)
		delete[]	pdString ;
	return	bRetval ;
}

BOOL
CSkkRuleTreeNode::bAddRule (
	CSkkRuleTreeNode**	ppTree,
	int					iTree,
	LPCDSTR				dstrPrefix,
	int					nPrefixLen,
	LPCWSTR				wstrNext,
	int					nNextLen,
	int					nNextTree,
	int					nMacro) 
{
	CSkkRuleTreeNode*		pRoot ;
	CSkkRuleTreeNode*		pAddPoint		= NULL ;
	CSkkRuleTreeNode*		pAddTree ;
	CSkkRuleTreeNode*		pAddTreeRoot	= NULL ;
	CSkkRuleTreeNodeOutput*	pOutput			= NULL ;
	LPCDSTR					pwPtr ;
	LPCDSTR					pwPtrEnd ;
	int						nCH, nOffset ;
	DCHAR					bufNext [32] ;
	DCHAR*					pdNext	= NULL ;
	int						nNextDLen ;
	BOOL					bRetval	= TRUE ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || dstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= pCreateInstance (iTree, 0, NULL, 0, NULL, 0, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}

	nOffset		= CSkkRuleTreeNode::iSearchTree (pRoot, dstrPrefix, nPrefixLen, &pAddPoint) ;

	pwPtr		= dstrPrefix + nOffset ;
	pwPtrEnd	= dstrPrefix + nPrefixLen ;
	if (pwPtr < pwPtrEnd) {
		CSkkRuleTreeNode*		pCurAddPoint		= NULL ;

		/*	S̐VKo^łꍇB*/
		/*
		 *	shu ƂāAs ݂ꍇɂ nOffset = 1B
		 *
		 *	[g -> ``s'' -> ...
		 *
		 *	 ``s''  pAddPoint ƂČtĂBA
		 *
		 *	(1) ``h'' -> ``u'' ̃N쐬B
		 *	(2) ``s'' ̎qƂĒǉB
		 *		ɂ ``s'' -> ``h'' -> ``u''  rule-tree 쐬B
		 *
		 *	 ``s'' ̎q݂ꍇɂ́A
		 *
		 *	(2)' ``h''->``u''  ``h'' ̌ZɂƂƂ ``s'' ̎qǉB
		 *	(3)  ``s'' ̎qƂ ``h'' ȂB
		 *
		 *	̃[ɈᔽȂ悤Ƀ[؂̍\gȂ΂ȂȂB
		 *	nCH  INT Ȃ̂ŁA̕ԈႦȂΊgɎs邱Ƃ͂Ȃ
		 *	̂ȂB
		 *	nCH ɓꕶꂽꍇ́uݏԂ̊ǗvB͂ǂɂ
		 *	Ȃ΂ȂȂB̓[؂̒ɓ̂͂܂ȂB
		 *	nCH ̓ꕶg
		 *		SelectBranch
		 *	͕ωB
		 *	Ԃ m_pPrefix, m_nPrefix A̎ۂۑꏊKvB
		 *	ۂ̏Ԃ́AiCH, m_pPrefix, m_nPrefix 쐬锤B
		 */
		pAddTreeRoot	= NULL ;
		while (pwPtr < pwPtrEnd) {
			LPCDSTR	pwPtrBack	= pwPtr ;

			nCH			= CSkkRuleTreeNode::iGetRuleCharacter (&pwPtr, pwPtrEnd) ;
			if (nCH == CSkkRuleTreeNode::SYSCHAR_INVALID) {
				return	FALSE ;
			}
			if (pwPtr >= pwPtrEnd) {
				pwPtr	= pwPtrBack ;
				break ;
			}
			pAddTree	= CSkkRuleTreeNode::pCreateInstance (iTree, nCH, dstrPrefix, pwPtr - dstrPrefix, NULL, 0, iTree, NULL) ;
			if (pAddTree == NULL) {
				vDestroyTree (pAddTreeRoot) ;
				return	FALSE ;
			}

			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			/*	addpoint  addtree ǉB\[gȂȂycB*/
			if (pAddTreeRoot == NULL) {
				pAddTreeRoot	= pAddTree ;
				pCurAddPoint	= pAddTree ;
			} else {
				pAddTree->vSetBrother (pCurAddPoint->pGetChild ()) ;
				pCurAddPoint->vSetChild (pAddTree) ;
				pCurAddPoint	= pAddTree ;
			}
		}

		/*	Output 쐬B
		 */
		pOutput		= CSkkRuleTreeNodeOutputMacro::pCreateInstance (nMacro) ;
		if (pOutput == NULL) {
			vDestroyTree (pAddTreeRoot) ;
			return	FALSE ;
		}

		/*	Output ̂Ȃ RuleTree ̃m[h쐬B
		 */
		nCH	= CSkkRuleTreeNode::iGetRuleCharacter (&pwPtr, pwPtrEnd) ;
		if (nCH == CSkkRuleTreeNode::SYSCHAR_INVALID) {
			vDestroyTree (pAddTreeRoot) ;
			delete	pOutput ;
			return	FALSE ;
		}

		pdNext		= pGetDString (bufNext, MYARRAYSIZE (bufNext), wstrNext, nNextLen, &nNextDLen) ;
		pAddTree	= CSkkRuleTreeNode::pCreateInstance (iTree, nCH, dstrPrefix, nPrefixLen, pdNext, nNextDLen, nNextTree, pOutput) ;
		if (pAddTree == NULL) {
			vDestroyTree (pAddTreeRoot) ;
			delete	pOutput ;
			bRetval	= FALSE ;
			goto	Exit ;
		}
		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		if (pAddTreeRoot != NULL) {
			ASSERT (pCurAddPoint != NULL && pCurAddPoint != pAddPoint) ;
			pAddTree->vSetBrother (pCurAddPoint->pGetChild ()) ;
			pCurAddPoint->vSetChild (pAddTree) ;
		} else {
			pAddTreeRoot	= pAddTree ;
		}

		/*	̒iKłƂ̖؂Ƀ}[WB
		 */
		pAddTreeRoot->vSetBrother (pAddPoint->pGetChild ()) ;
		pAddPoint->vSetChild (pAddTreeRoot) ;
	} else {
		pOutput		= CSkkRuleTreeNodeOutputMacro::pCreateInstance (nMacro) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddPoint->vSetOutput (pOutput) ;
	}
Exit:
	if (pdNext != bufNext)
		delete[]	pdNext ;
	return	bRetval ;
}

BOOL
CSkkRuleTreeNode::bDeleteRule (
	CSkkRuleTreeNode**	ppTree,
	LPCDSTR				pdString,
	int					nString)
{
	CSkkRuleTreeNode*	pTree ;
	CSkkRuleTreeNode*	pParent ;
	CSkkRuleTreeNode*	pNode ;
	CSkkRuleTreeNode*	pPrev ;
	LPCDSTR		wptr ;
	int			nptr, nCH ;

	pTree	= *ppTree ;
	pNode	= pTree != NULL? pTree->pGetChild () : NULL ;
	pParent	= pTree ;
	wptr	= pdString ;
	nptr	= nString ;

	/*	܂ brother ւ̒TsB*/
	while (pNode != NULL && nptr > 0) {
		nCH		= *wptr ;
		pPrev	= NULL ;
		while (pNode != NULL) {
			if (pNode->iGetChar () == nCH) 
				break ;
			pPrev	= pNode ;
			pNode	= pNode->pGetBrother () ;
		}
		/* YȂB*/
		if (pNode == NULL)
			return	FALSE ;

		wptr	++ ;
		nptr	-- ;
		if (nptr <= 0) {
			/* ̈ʒuō폜B*/
			if (pPrev == NULL) {
				/* ẽN\ւB*/
				if (pParent != NULL) {
					pParent->vSetChild (pNode->pGetBrother ()) ;
				} else {
					*ppTree	= pNode->pGetBrother () ;
				}
				pNode->vSetBrother (NULL) ;
			} else {
				/*	Z̃N\ւB
				 *	cqq͑SłB
				 */
				pPrev->vSetBrother (pNode->pGetBrother ()) ;
				pNode->vSetBrother (NULL) ;
			}
			CSkkRuleTreeNode::vDestroyTree (pNode) ;
			break ;
		}
		pParent	= pNode ;
		pNode	= pNode->pGetChild () ;
	}
	return	TRUE ;
}

void
CSkkRuleTreeNode::vDestroyTree (
	CSkkRuleTreeNode*	pRoot)
{
	CSkkRuleTreeNode*	pBrother ;
	CSkkRuleTreeNode*	pChild ;

	if (pRoot == NULL)
		return ;

	pBrother	= pRoot->pGetBrother () ;
	pChild		= pRoot->pGetChild () ;
	pRoot->vSetBrother (NULL) ;
	pRoot->vSetChild (NULL) ;
	delete	pRoot ;

	if (pBrother != NULL)
		CSkkRuleTreeNode::vDestroyTree (pBrother) ;
	if (pChild != NULL)
		CSkkRuleTreeNode::vDestroyTree (pChild) ;
	return ;
}

int
CSkkRuleTreeNode::iGetRuleCharacter (
	LPCDSTR*	ppdPtr,
	LPCDSTR		pdPtrEnd)
{
	LPCDSTR	pwPtr ;
	int		nCH ;

	if (ppdPtr == NULL || pdPtrEnd == NULL)
		return	CSkkRuleTreeNode::SYSCHAR_INVALID ;

	pwPtr	= *ppdPtr ;
	if (pwPtr == NULL)
		return	CSkkRuleTreeNode::SYSCHAR_INVALID ;

	if (*pwPtr == L'\\') {
		pwPtr	++ ;
		if (pwPtr >= pdPtrEnd) {
			nCH	= CSkkRuleTreeNode::SYSCHAR_INVALID ;
		} else {
			switch (*pwPtr) {
			case '\\':
				nCH	= '\\' ;
				break ;
			case	'V':	/* Vowel */
				nCH	= CSkkRuleTreeNode::SYSCHAR_VOWEL ;
				break ;
			case	'C':
				nCH	= CSkkRuleTreeNode::SYSCHAR_CONSONANT ;
				break ;
			case	'?':
				nCH	= CSkkRuleTreeNode::SYSCHAR_DEFAULT ;
				break ;
			default:
				nCH	= CSkkRuleTreeNode::SYSCHAR_INVALID ; 
				break ;
			}
		}
	} else {
		nCH	= *pwPtr ;
	}
	pwPtr	++ ;
	*ppdPtr	= pwPtr ;
	return	nCH ;
}

/*===============================================================
 *	public methods
 */
#if 0
void
CSkkRuleTreeNode::vDebugOut () const 
{
	DCHAR	buf [256] ;
	DCHAR	bufPrefix [128], bufNextState [128] ;
	int		n ;

	if (m_pPrefix != NULL && m_nPrefix > 0) {
		n	= (m_nPrefix < (MYARRAYSIZE (bufPrefix) - 1))? m_nPrefix : (MYARRAYSIZE (bufPrefix) - 1) ;
		memcpy (bufPrefix, m_pPrefix, n * sizeof (DCHAR)) ;
		bufPrefix [n]	= L'\0' ;
	} else {
		bufPrefix [0]	= L'\0' ;
	}
	if (m_pNextState != NULL && m_nNextState > 0) {
		n	= (m_nNextState < (MYARRAYSIZE (bufNextState) - 1))? m_nNextState : (MYARRAYSIZE (bufNextState) - 1) ;
		memcpy (bufNextState, m_pNextState, n * sizeof (DCHAR)) ;
		bufNextState [n]	= L'\0' ;
	} else {
		bufNextState [0]	= L'\0' ;
	}
	if (m_pOutput != NULL) {
		switch (m_pOutput->iGetType ()) {
		case	CSkkRuleTreeNodeOutput::NTYPE_STRINGPAIR:
			{
				CSkkRuleTreeNodeOutputStringPair*	pOutput	= static_cast <CSkkRuleTreeNodeOutputStringPair*> (m_pOutput) ;
				LPCDSTR	pHira, pKata ;
				int		nHira, nKata ;
				DCHAR	bufHira [128], bufKata [128] ;

				pHira	= pOutput->pGetHiragana (&nHira) ;
				pKata	= pOutput->pGetKatakana (&nKata) ;
				if (pHira != NULL && nHira > 0) {
					n	= (nHira < (MYARRAYSIZE (bufHira) - 1))? nHira : (MYARRAYSIZE (bufHira) - 1) ;
					memcpy (bufHira, pHira, nHira * sizeof (DCHAR)) ;
					bufHira [n]	= L'\0' ;
				} else {
					bufHira [0]	= L'\0' ;
				}
				if (pKata != NULL && nKata > 0) {
					n	= (nKata < (MYARRAYSIZE (bufKata) - 1))? nKata : (MYARRAYSIZE (bufKata) - 1) ;
					memcpy (bufKata, pKata, nKata * sizeof (DCHAR)) ;
					bufKata [n]	= L'\0' ;
				} else {
					bufKata [0]	= L'\0' ;
				}
				wnsprintfW (buf, MYARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), Hira(\"%s\"), Kata(\"%s\")\n", bufPrefix, bufNextState, bufHira, bufKata) ;
				buf [MYARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;

		case	CSkkRuleTreeNodeOutput::NTYPE_STRING:
			{
				CSkkRuleTreeNodeOutputString*	pOutput	= static_cast <CSkkRuleTreeNodeOutputString*> (m_pOutput) ;
				DCHAR	bufString [128] ;
				LPCDSTR	pString ;
				int		nString ;

				pString	= pOutput->pGetString (&nString) ;
				if (pString != NULL && nString > 0) {
					n	= (nString < (MYARRAYSIZE (bufString) - 1))? nString : (MYARRAYSIZE (bufString) - 1) ;
					memcpy (bufString, pString, nString * sizeof (DCHAR)) ;
				} else {
					bufString [0]	= L'\0' ;
				}
				wnsprintfW (buf, MYARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), String(\"%s\")\n", bufPrefix, bufNextState, bufString) ;
				buf [MYARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;

		case	CSkkRuleTreeNodeOutput::NTYPE_MACRO:
			{
				CSkkRuleTreeNodeOutputMacro*	pOutput	= static_cast <CSkkRuleTreeNodeOutputMacro*> (m_pOutput) ;
				int		nMacro ;

				nMacro	= pOutput->iGetFunctionCode () ;
				wnsprintfW (buf, MYARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), Macro(%d)\n", bufPrefix, bufNextState, nMacro) ;
				buf [MYARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;
		default:
			break ;
		}
		wprintf (L"%s\n", buf) ;
	} else {
		wprintf (L"nil\n", buf) ;
	}
	return ;
}
#endif

#if 0 //defined (UNIT_TEST)
/*	P̃eXĝ߂̃R[hB
 */
int
_tmain (void)
{
	CSkkRuleTreeNode*	pRoot	= NULL ;
	CSkkRuleTreeNode*	pNode	= NULL ;
	int		i, n ;
	char	buf [12] ;

	/*
struct TSkkBaseRuleT1 {
	LPCTSTR	_strState ;
	LPCTSTR	_strNext ;
	struct {
		LPCTSTR	_strHira ;
		LPCTSTR	_strKata ;
	}	_case ;
} ;*/
	_tsetlocale (LC_CTYPE, TEXT ("")) ;

	/* debugger  process  attach 邽߂̗]TB*/
	_tprintf (TEXT ("> ")) ;
	fflush (stdout) ;
	fgets (buf, 10, stdin) ;

	for (i = 0 ; i < MYARRAYSIZE (_rT1) ; i ++) {
		int	nPrefix, nNextState, nHira, nKata ;

		nPrefix		= (_rT1 [i]._strState      != NULL)? lstrlenW (_rT1 [i]._strState)      : 0 ;
		nNextState	= (_rT1 [i]._strNext       != NULL)? lstrlenW (_rT1 [i]._strNext)       : 0 ;
		nHira		= (_rT1 [i]._case._strHira != NULL)? lstrlenW (_rT1 [i]._case._strHira) : 0 ;
		nKata		= (_rT1 [i]._case._strKata != NULL)? lstrlenW (_rT1 [i]._case._strKata) : 0 ;
		if (! TSkkRuleTreeNode_bAddRule (&pRoot, _rT1 [i]._strState, nPrefix, _rT1 [i]._strNext, nNextState, _rT1 [i]._case._strHira, nHira, _rT1 [i]._case._strKata, nKata)) {
			_tprintf (TEXT ("TSkkRuleTreeNode_bAddRule: failed. (%d)\n"), i) ;
			break ;
		}
	}
	if (pRoot != NULL) {
		_tprintf (TEXT ("Root ...\n")) ;
		TSkkRuleTreeNode_vDebugOut (pRoot) ;
	}
	n	= TSkkRuleTreeNode_iSearchTree (pRoot, L"b", 1, &pNode) ;
	_tprintf (TEXT ("Search ... \"b\" ... hit(%d)\n"), n) ;
	if (pNode != NULL) {
		TSkkRuleTreeNode_vDebugOut (pNode) ;
	}
	n	= TSkkRuleTreeNode_iSearchTree (pRoot, L"fya", 3, &pNode) ;
	_tprintf (TEXT ("Search ... \"fya\" ... hit(%d)\n"), n) ;
	if (pNode != NULL) {
		TSkkRuleTreeNode_vDebugOut (pNode) ;
	}
	return	0 ;
}
#endif



